@rem = '-*- Perl -*-';
@rem = '
@echo off
perl %0.cmd %1 %2 %3 %4 %5 %6 %7 %8 %9
goto endofperl
';
# DskLayt2.Perl (version 1.0)				 8 July 1993
#
# Perl script to replace Microsoft's DSKLAYT2 utility for Setup
# Splits and compresses files, assigns them to disks, and creates .INF file
#
# Created by:
#     Michael Longe'
#     Designer Software
#     P.O. Box 4502
#     Federal Way, WA  98063
#     (206) 946-5218
# for Statistical Sciences, Inc. (now a division of MathSoft, Inc.)
#
# Modified to deal with the 32-bit version of MS Setup by:
#     Clark Williams
#     Intergraph Corporation
# 
# Copyright (C) 1993 Michael Longe'
# Permission is granted to freely modify and redistribute the source and
# accompanying documentation provided this authorship and copyright
# information is retained.
#
# Copyright (C) 1993 Clark Williams
# Permission is granted to freely modify and redistribute the source and
# accompanying documentation provided this authorship and copyright
# information is retained.
#
# Hazards to avoid in MS-DOS version:
#   system() requires explicit $ENV{'COMSPEC'} /c call in most cases
#   COPY won't create subdirectories or copy zero-length files; XCOPY asks 
#     for clarification on every file!
#   must replace / with \ in paths before exporting them, except for args
#     to MKS Toolkit utilities
#

# global variables and defines
@disk_labels = ();
$disk_size = 1185 * 2;			# default size is 1.2MB
$copy_list_count = 0;			# local to &addcopylist
%copy_list = ();
$overwrite_inf = 0;
$asis = 0;
$group = 0;
$verbose = 0;
$quiet = 0;
$comspec = $ENV{'COMSPEC'} if defined($ENV{'COMPSPEC'});
$comspec = $ENV{'ComSpec'} if !defined($compspec);
$split_id = "AA";		# starting id for each split file
$any_order = "ZZz";		# if assign to any disk # will do
$flatten = 0;			# flatten delivery directory

# check index of last arg (should be at least 2 args)
if ($#ARGV < 1 || ! -e $ARGV[0])
{
    &usage();
    exit (1);
}
# process command-line options (- or / OK)
@args = splice (@ARGV, 2, $#ARGV - 1);
while ($args[0] =~ /^[\/-]/)
{
	$_ = shift @args;
	if (/^.k([0-9]+)/)
	{
	    # size of dest disk in KB specified
	    # (should also handle standard sizes)
	    $disk_size = $1 * 2;		# size in 512-byte blocks 
	}
	elsif (/^.kN/) {
	    $netdeliv = 1;
	    $disk_size = 1024 * 1024 * 1024; # one gigabyte
	}
	elsif (/^.d/) {
	    s/^.d *//;
	    $destdir = length($_) ? $_ : shift @args;
	}
	elsif (/^.F/) {
	    $flatten = 1;
	}
	elsif (/^.f/)
	{
	    # overwrite .INF file specified
	    $overwrite_inf = 1;
	}
	elsif (/^.c(\S+)/)
	{
	    # dir for compression files specified
	    $comp_dir = $1;
	    $comp_dir =~ s,\\,/,go;
	    $comp_dir .= "/"  unless $comp_dir =~ /\/$/;
	}
	elsif (/^.z"?([^"]+)"?/)
	{
	    # command line for compression specified
	    $comp_cmd = $1;
	}
	elsif (/^.v/)
	{
	    # verbose mode specified
	    $verbose = 1;
	}
	elsif (/^.a/)
	{
	    # accept split files as-is specified
	    $asis = 1;
	}
	elsif (/^.g([sSrR])([0-9]*)/)
	{
	    # group assignments specified
	    local($opt) = $1;
	    $group_keyw = (defined ($2) && length ($2) > 0) ? $2 : 0;
	    $group = ($opt =~ /[sS]/o) ? 1 : 2;
	}
	elsif (/^.s(\S+)/)
	{
	    # dir for split files specified
	    $split_dir = $1;
	    $split_dir =~ s,\\,/,go;
	    $split_dir .= "/"  unless $split_dir =~ /\/$/;
	}
	elsif (/^.t(\S+)/)
	{
	    # temp dir for split files specified
	    $temp_dir = $1;
	    $temp_dir =~ s,\\,/,go;
	    $temp_dir .= "/"  unless $temp_dir =~ /\/$/;
	}
	elsif (/^.q/)
	{
	    # quiet mode (nothing to STDOUT) specified
	    $quiet = 1;
	}
	else
	{
	    print STDERR "Unrecognized option: '$_'\n";
	    &usage ();
	    exit (1);
	}
}
$verbose = 0  if $quiet;

# open input and output files
unless (open(LYTFILE, $ARGV[0]))
{
    print STDERR "Can't open file '$ARGV[0]'\n";
    exit(1);
}
if (-e $ARGV[1] && !$overwrite_inf)
{
    print STDERR "File '$ARGV[1]' exists; use -f option\n";
    exit(1);
}
unless (open(INFFILE, ">" . $ARGV[1]))
{
    print STDERR "Can't open file '$ARGV[1]' for writing\n";
    exit(1);
}
$INFFILE = $ARGV[1];

# process all header lines
print STDOUT "Reading layout file\n"  if $verbose;
$disk_num = 0;
while (<LYTFILE>)
{
    chop $_;
    if (/^SRC\s*=\s*(\S.*)$/)
    {
	# found SRC line; switch \ to / also
	$src_dir = $1;
	$src_dir =~ s,/,\\,go;
	$src_dir .= "\\"  unless $src_dir =~ /\\$/;
    }
    elsif (/^DISKLABEL\s*=\s*(\S.*)$/)
    {
	# found DISKLABEL line; save names for later
	$disk_num++;
	push (@disk_labels, join('", "', $disk_num, $1, "#fn#", ""));
    }
    elsif (/^FILES/)
    {
	# found FILES list; go to next section
	last;
    }
    # do anything with others?
}

# set default dirs if necessary
chop($cwd = `cd`);
($cur_drive = $cwd) =~ s/^([a-z]:).*$/$1/i;
$src_dir = "./"  if !defined($src_dir);
$comp_dir = $src_dir . "..\\comp\\"  if !defined($comp_dir);
$split_dir = $comp_dir  if !defined($split_dir);
$comp_cmd = "compress.exe"  if !defined($comp_cmd);

# use Windows TEMP env variable, if available
$temp_dir = $ENV{'TEMP'}  if (!defined($temp_dir) && defined($ENV{'TEMP'}));
$temp_dir = $ENV{'TMP'} if (!defined($temp_dir) && defined($ENV{'TMP'}));
$temp_dir = $ENV{'Temp'} if (!defined($temp_dir) && defined($ENV{'Temp'}));
$temp_dir = $ENV{'Tmp'} if (!defined($temp_dir) && defined($ENV{'Tmp'}));
$temp_dir = $ENV{'temp'} if (!defined($temp_dir) && defined($ENV{'temp'}));
$temp_dir = $ENV{'tmp'} if (!defined($temp_dir) && defined($ENV{'tmp'}));

if (defined($temp_dir) && length($temp_dir) > 0) {
    $temp_dir =~ s,/,\\,go;
    $temp_dir .= "\\"  unless $temp_dir =~ /\\$/;
}
else {
    $temp_dir = "$cur_drive\\tmp\\";
    # create if dir doesn't exist
    &makedirs ($temp_dir . "dummy");
}

# process all filelist lines
print STDOUT "Processing layout file list ...\n"  unless $quiet;
while (<LYTFILE>) {
    if (/^REMOVE/) {
	# assume REMOVE always marks the last line
	last;
    }

    chop $_;
    s,/,\\,go;	# change any forward slashes to back slashes

    # split into fields (there may be more instead of the period)
    ($fname, $disk, $comp, $ver, $read, $date, $rename, $owrite, $back, $decomp, $vital, $sect, $ref, $sys, $share, $per) = split(/,/, $_);

    $src_fpath = $src_dir . $fname;
    if (! -e $src_fpath) {
	# ERROR: source file not present
	print STDERR "Can't open '$src_fpath'\n";
	exit (1);
    }

    #
    # if requested, flatten the deliv directory. 
    # This is usually done because the setup procedure doesn't 
    # know how to deal with subdirectories.
    #
    
    if ($flatten) {
	@tmp = split (/\\/, $fname);
	$fname = pop(@tmp);
    }

    #$fname =~ s/.$/_/ if $comp eq "COMPRESS";

    # other per-file variables 
    $size = -s $src_fpath;
    $src_age = -M $src_fpath;

    # revise format of a few fields
    $sect =~ s/SECTION=//o;		# explicit section or nothing
    $ref =~ s/REFKEY=//o;		# explicit ref key or nothing
    $read =~ s/^\s*READONLY//o;		# only explicit !READONLY

    # use explicit date or get date from file
    if (!($date =~ s/DATE=//o)) {
	# stat source file, convert mod date, format as YYYY-MM-DD
	local($sd,$si,$sm,$sl,$su,$sg,$sr,$ss,$sat,$smt,$sct,$sbs,$sb) = 
	    stat ($src_fpath);
	local($gs,$gm,$gh,$gmd,$gm,$gy,$gwd,$gyd,$gid) = gmtime ($smt);

	$date = sprintf ("%4d-%02d-%02d", $gy+1900, $gm+1, $gmd);
    }

    # compare src mod age against comp mod age
    if ($comp eq "COMPRESS") {
	$comp_fpath = $comp_dir . $fname;
	$comp_age = -M $comp_fpath;
	if (! -e $comp_fpath || $comp_age > $src_age) {
	    # compress it again, Sam
	    &compress ($src_fpath, $comp_fpath);
	    $comp_age = -M $comp_fpath;
	    if (! -e $comp_fpath || $comp_age > $src_age) {
		# ERROR: compress must have failed
		print STDERR "Compression of '$src_fpath' failed\n";
		exit (2);
	    }
	}
	# check size of compressed file against disk size
	$comp_size = -s $comp_fpath;
	$block_size = int (($comp_size + 511) / 512);
	if ($block_size > $disk_size) {
	    # split file and try again (assume only 2 parts needed)
	    local($split_fpath) = $split_dir . $fname;
	    local($fname1, $fname2) = &splitname ($fname, 2);
	    local($split_fpath1, $split_fpath2) = &splitname ($split_fpath, 2);

	    # check existence and age if $asis
	    if ($asis && -e $split_fpath1 
		&& -M $split_fpath1 < $comp_age && 
		-e $split_fpath2 && -M $split_fpath2 < $comp_age) {
		print STDOUT "  Accepting $split_fpath1, $split_fpath2 as-is\n"
		    unless $quiet;
		# add two comp files to list; first is renamed
		$block_size = int (((-s $split_fpath1) + 511) / 512); 
		&addcopylist ($split_id . "1", $block_size, $split_fpath1, 
			      $sect, $disk, $fname1, $disk_size * 512, "", 
			      $read, $ref, $date, "", $fname);
		# second file is appended to first
		$block_size = int (((-s $split_fpath2) + 511) / 512); 
		&addcopylist ($split_id . "2", $block_size, $split_fpath2, 
			      $sect, $disk, $fname2, 
			      $size - ($disk_size * 512), "", $read, 
			      $ref, $date, $fname, "");
		$split_id++;		# magic auto-incr "AA" to "AB", etc.
		next;
	    }

	    print STDOUT "  Splitting $fname into $fname1, $fname2\n" 
		unless $quiet;
	    local($part1, $part2) = &splitfile ($src_fpath, $disk_size);
	    &compress ($part1, $split_fpath1);
	    &compress ($part2, $split_fpath2);
	    $block_size = int (((-s $split_fpath1) + 511) / 512); 
	    if ($block_size > $disk_size) {
		# ERROR: too big to fit
		print STDERR "Compressed '$split_fpath1' too large to fit on any disk\n";
		exit (2);
	    }
	    # add two comp files to list; first is renamed
	    &addcopylist ($split_id . "1", $block_size, $split_fpath1, 
			  $sect, $disk, $fname1, -s $part1, "", $read, 
			  $ref, $date, "", $fname);
	    $block_size = int (((-s $split_fpath2) + 511) / 512); 
	    if ($block_size > $disk_size) {
		# ERROR: too big to fit
		print STDERR "Compressed '$split_fpath2' too large to fit on any disk\n";
		exit (2);
	    }
	    # second file is appended to first
	    &addcopylist ($split_id . "2", $block_size, $split_fpath2, 
			  $sect, $disk, $fname2, -s $part2, "", $read, 
			  $ref, $date, $fname, "");
	    unlink ($part1, $part2);
	    $split_id++;		# magic auto-incr "AA" to "AB", etc.
	    next;
	}

	# add comp file to list (no append, no rename)
	&addcopylist ($any_order, $block_size, $comp_fpath, $sect, 
		      $disk, $fname, $size, "", $read, $ref, $date, "", "");
	next;
    }

    # otherwise, add original file to list
    $block_size = int (($size + 511) / 512);
    if ($block_size > $disk_size) {
	# should split original file into parts and put in split dir
	# check existence and dates here if $asis
	# ERROR: too big to fit
	print STDERR "Non-compressed '$src_fpath' too large to fit on any disk\n";
	exit (2);
    }
    &addcopylist ($any_order, $block_size, $src_fpath, $sect, $disk, 
		  $fname, $size, "!DECOMPRESS", $read, $ref, $date, "", "");
}

print STDOUT "Processing complete.\n"  unless $quiet;

if (!defined($netdeliv)) {
    print STDOUT "Assigning files to disks ...\n"  if $verbose;
    
# copylist is an associative array with order/size/id as index
# iterate through, building full disks, skipping entries with assigned disks
# (sure would be nice if DSKLAYT allowed disks other than 1 to be specified,
# or at least a grouping to keep related files together)
    @keys = sort keys (%copy_list);

# split_ids is an associative array with split id as index
# the value indicates which split file is to be assigned next
    %split_ids = ();
    foreach $id ("AA" .. $split_id) {
	$split_ids{$id} = 1;
    }

# now go through each disk (skipping #1 -- assume it's the Setup disk)
    $disk_labels[0] =~ s/#fn#/setup.exe/;
    $disk_num = 1;
  DISK_LOOP:
    do {
	$disk_num++;
	$disk_total = 0;
	$unassigned = 0;
	
	foreach $key (@keys) {
	    local($sect, $disk, $fname, $block_size, @etc) = 
		split(/,/, $copy_list{$key});
	    
	    if ($disk eq "#x#") {
		# found an unassigned entry; try to fit it
		$unassigned++;
		next  if ($block_size + $disk_total > $disk_size);
		
		# see if it is in order (key: "<group>!<file id><order>...")
		# NOTE: file groups are not enforced (small files from secondary
		# groups may be on an earlier disk than larger files from primary)
		$key =~ /!(..)(.)/o;
		local ($id, $order) = ($1, $2);
		next  if ($id . $order ne $any_order && $split_ids{$id} != $order);
		
		# will fit; assign current disk number
		$copy_list{$key} =~ s/#x#/$disk_num/;
		$disk_total += $block_size;
		# also mark disk label with first file assigned
		$fname =~ s,/,\\,go;
		$disk_labels[$disk_num - 1] =~ s/#fn#/$fname/;
		print STDOUT "  Assigning $fname to disk $disk_num\n"  if $verbose;
		
		# increment next file in order, if any
		$split_ids{$id}++  if ($split_ids{$id} == $order);
		$unassigned--;
	    }
	}
    } until (!$unassigned);
    print STDOUT "Assignments complete.\n"  if $verbose;
}

# now sort the copylist by section, disk #, and filename
@sorted_list = sort (values (%copy_list));

print STDOUT "Creating .INF file ...\n"  if $verbose;

@disk_labels = ('"1", "disk1", "setup.exe", "."') if $netdeliv;

# stuff standard file header into INFFILE
&copy_header(INFFILE);

# and output each .INF line
$section = "";
foreach $entry (@sorted_list) {
    local($sect, $disk, $fname, 
	  $block_size, $src_fpath, $size, $decomp, 
	  $read, $date, $refkey, $append, $rename) = split(/,/, $entry);
    if ($section ne $sect) {
	$section = $sect;
	print INFFILE "\n[$section]\n";
    }
    print INFFILE "\t";
    print INFFILE "\"$refkey\" = "  if length($refkey) > 0;

    print INFFILE 
	"$disk,$fname,$append,,,$date,$decomp,,,$read,,$rename,,,,$size,,,,,\n";
}

&copy_footer (INFFILE);

close (LYTFILE);
close (INFFILE);
print STDOUT ".INF file complete.\n"  if $verbose;



if ($netdeliv) {
    die "No destination for network delivery!" if !defined($destdir);    
    &do_file_copy($destdir);
}
 
print STDOUT "dsklayt2 finished.\n"  unless $quiet;

# end of dsklayt2 main processing

sub addcopylist {
    local($order, $block_size, $cur_path, $sect, $disk, $fname, $size, 
	  $decomp, $read, $refkey, $date, $append, $rename) = @_;
    local($copy_key, $group_key);

    # build key & value to copy_list
    $copy_list_count++;
    $disk = ($disk eq "DISK1") ? "1" : defined($netdeliv) ? "1" : "#x#";
    if ($group == 1) {
	# group by section; strip off key, if any
	if ($group_keyw) {
	    $group_key = substr ($sect, 0, $group_keyw);
	    $sect = substr ($sect, $group_keyw);
	}
	else {
	    $group_key = $sect;
	}
    }
    elsif ($group == 2) {
	# group by reference key; strip off key, if any
	if ($group_keyw) {
	    $group_key = substr ($refkey, 0, $group_keyw);
	    $refkey = substr ($refkey, $group_keyw);
	}
	else {
	    $group_key = $refkey;
	}
    }

    # key format: "<group>!<split file id/order><size><serial #>"
    # the group key is optional; the split file id & order ensures that a 
    # file is reassembled in the correct order; the size in blocks (subtracted
    # for max-to-min sort) allows best-fit disk packing; serial # guarantees
    # unique key
    # this key ordering makes split files a priority; if other priorities
    # are desired, the key parts can be rearranged
    $copy_key = sprintf ("%s!%s%4.4lx%3.3x", $group_key, $order, 
			 0xffff - $block_size, $copy_list_count);
    $copy_list{$copy_key} = join (',', $sect, $disk, $fname, 
				  $block_size, $cur_path, $size, $decomp, 
				  $read, $date, $refkey, $append, $rename);
}

sub splitfile {
    local($fpath, $max_blocks) = @_;
    local(@parts) = ();
    local($block_size) = int (((-s $fpath) + 511) / 512); 
    local($num_parts) = int (($block_size + $max_blocks - 1) / $max_blocks);
    local($block_off) = 0;

    return if defined($netdeliv);

    # divide file into 2-9 parts, each less than or equal to max_blocks
    foreach $i (1 .. $num_parts) {
	local($part) = $temp_dir . "part$i.tmp";

	push (@parts, $part);
	# change / to \ as necessary (NOT with MKS "dd")
	#$part =~ s,/,\\,go;
	#$fpath =~ s,/,\\,go;

	# use "dd" here if available
	if ($i == 1) {
	    system ($comspec, "/c dd if=$fpath of=$part bs=512 count=$max_blocks");
	    #UNIX: system ("dd if=$fpath of=$part bs=512 count=$max_blocks 2>/dev/null");
	}
	else {
	    system ($comspec, "/c dd if=$fpath of=$part bs=512 iseek=$block_off count=$max_blocks");
	    #UNIX: system ("dd if=$fpath of=$part bs=512 skip=$block_off count=$max_blocks 2>/dev/null");
	}
	$block_off += $max_blocks;
    }

    #return filenames
    @parts;
}

sub splitname {
    local($fname, $count) = @_;
    local(@parts) = ();
    local($pos, $i);

    return if defined($netdeliv);

    # create names which are similar to original but have last 2 chars unique
    # (9 parts max)

    $pos = rindex ($fname, ".");
    if ($pos == -1 || rindex ($fname, "\\") > $pos) {
	# no ext; append a dot
	$fname .= ".";
	$pos = length ($fname);
    }
    else {
	# step over dot, and char following if any
	$pos++;
	$pos++ if ($pos < length ($fname));
    }
    $i = 1;
    foreach $i (1 .. $count) {
	substr ($fname, $pos) = $i . "_";
	push (@parts, $fname);
    }

    #return filenames
    @parts;
}

sub copy_header {
    local($fd) = @_;
    local($entry,@items);

    print $fd "[Source Media Descriptions]\n";

    foreach $entry (@disk_labels) {
	print $fd "\t", $entry, "\n";
    }

    # change these to match best defaults
    print $fd "\n", '[Default File Settings]', "\n";
    print $fd '"STF_BACKUP"     = ""', "\n";
    print $fd '"STF_COPY"       = "YES"', "\n";
    print $fd '"STF_DECOMPRESS" = "YES"', "\n";
    print $fd '"STF_OVERWRITE"  = "ALWAYS"', "\n";
    print $fd '"STF_READONLY"   = ""', "\n";
    print $fd '"STF_ROOT"       = ""', "\n";
    print $fd '"STF_SETTIME"    = "YES"', "\n";
    print $fd '"STF_TIME"       = ""', "\n";
    print $fd '"STF_VITAL"      = "YES"', "\n";
}

sub copy_footer {
    local($fd) = @_;

}

sub makedirs {
    local($path) = @_;
    local(@dirs);

    # create each subdir along the path as necessary

    @dirs = split(/\\/, $path);

    # assume last name is always a file
    return  if ($#dirs == 0);

    # starts with first (dir), never does last (file)
    $path = shift (@dirs);
    foreach $dir (@dirs)
    {
	if ($path !~ /^[a-z]:$/i) {
	    unless (-e $path) {
		mkdir ($path, 0777) || 
		    die "can't create directory $path: $!\n";
	    }
	}
	$path .= "\\" . $dir;
    }
}

sub usage {
    print STDERR "Usage: dsklayt2 layout-file inf-file [options]\n";
    print STDERR "       layout-file         pathname for .LYT file\n";
    print STDERR "       inf-file            pathname for .INF file (output)\n";
    print STDERR "   options:\n";
    print STDERR "       -f                  overwrite .INF file, if necessary\n";
    print STDERR "       -F                  flatten delivery directory\n";
    print STDERR "       -k{disk-size}       size of destination disks, in KB\n";
    print STDERR "       -c{comp-dir}        compressed files directory (..\\comp default)\n";
    print STDERR "       -z{[\"]comp-cmd[\"]}  compression command\n";
    print STDERR "       -v                  verbose mode\n";
    print STDERR "   new options:\n";
    print STDERR "       -a                  accept split files as-is\n";
    print STDERR "       -g{s|r}{key-width}  group by Section or Ref; width of prefix to use\n";
    print STDERR "       -s{split-dir}       split file directory (comp-dir default)\n";
    print STDERR "       -t{temp-dir}        temporary files directory (TEMP or \\tmp default)\n";
    print STDERR "       -q                  quiet mode (turns off all output)\n";
}

sub do_file_copy {
    local($dir) = @_;
    local(@dirs);
    local($tmp);

    print STDOUT "Copying files to $dir\n" if $verbose || !$quiet;
    &makedirs($dir) if ! -d $dir;

    foreach (@sorted_list) {
	($sect, $disk, $fname, $block_size, $cur_path, $size, $decomp, 
	 $read, $date, $refkey, $append, $rename) = split (',',$_);
	if ($flatten == 0) {
	    @dirs = split(/[\/\\]/, $fname);
	    pop(@dirs);
	    $tmp = "$dir\\" . join('\\', @dirs);
	    &makedirs ($tmp) if ! -d $tmp;
	}
	$tmp = "$dir\\$fname";
	
	&copy ($cur_path, $tmp);
    }
    &copy ($INFFILE, "$dir\\$INFFILE");
}

sub copy {
    local ($src, $dst) = @_;
    local (@dirs) = split (/\\/, $dst);
    local ($cmd) = "copy $src $dst";
    local ($tmp) = $dst;

    &makedirs ($dst);
    print "$cmd\n" if $verbose;
    system "$comspec /c $cmd >nul:";
    die "Couldn't execute $cmd: $!" if $? >> 8;
}

sub compress {
    local($src, $dest) = @_;
    local($cmd) = "$comp_cmd $src $dest";

    # create subdirs as necessary
    &makedirs ($dest);

    # change / to \ as necessary

    print STDOUT "$cmd\N" if $verbose;

    system "$cmd";
    die "$cmd failed: $!" if $? >> 8;

}
__END__
:endofperl
